@@ -364,22 +364,26 @@ Call* aggregate_and_or_ops(EnvI& env, BinOp* bo, bool negateArgs, BinOpType bot)
     auto* bo_arg = i->first->dynamicCast<BinOp>();
     UnOp* uo_arg = i->first->dynamicCast<UnOp>();
     bool positive = i->second;
-    if ((bo_arg != nullptr) && positive && bo_arg->op() == bot) {
+    bool isBuiltin =
+        (bo_arg != nullptr && (bo_arg->decl() == nullptr || bo_arg->decl()->e() == nullptr)) ||
+        (uo_arg != nullptr && (uo_arg->decl() == nullptr || uo_arg->decl()->e() == nullptr));
+    if ((bo_arg != nullptr) && positive && bo_arg->op() == bot && isBuiltin) {
       i->first = bo_arg->lhs();
       i++;
       bo_args.insert(i, arg_literal(bo_arg->rhs(), true));
       i--;
       i--;
-    } else if ((bo_arg != nullptr) && !positive && bo_arg->op() == negbot) {
+    } else if ((bo_arg != nullptr) && !positive && bo_arg->op() == negbot && isBuiltin) {
       i->first = bo_arg->lhs();
       i++;
       bo_args.insert(i, arg_literal(bo_arg->rhs(), false));
       i--;
       i--;
-    } else if ((uo_arg != nullptr) && !positive && uo_arg->op() == UOT_NOT) {
+    } else if ((uo_arg != nullptr) && !positive && uo_arg->op() == UOT_NOT && isBuiltin) {
       i->first = uo_arg->e();
       i->second = true;
-    } else if (bot == BOT_OR && (uo_arg != nullptr) && positive && uo_arg->op() == UOT_NOT) {
+    } else if (bot == BOT_OR && (uo_arg != nullptr) && positive && uo_arg->op() == UOT_NOT &&
+               isBuiltin) {
       output_neg.push_back(uo_arg->e());
       i++;
     } else {
@@ -1016,7 +1020,7 @@ bool flatten_dom_constraint(EnvI& env, Ctx& ctx, VarDecl* vd, Expression* dom, V
         } else if (changeDom) {
           set_computed_domain(env, ident->decl(), new SetLit(Location().introduce(), newdom),
                               false);
-          if (ident->decl()->e() == nullptr && newdom->min() == newdom->max() &&
+          if (ident->decl()->e() == nullptr && !newdom->empty() && newdom->min() == newdom->max() &&
               !vd->type().isSet()) {
             ident->decl()->e(LinearTraits<Lit>::newLit(newdom->min()));
           }
@@ -1204,9 +1208,30 @@ EE flatten_bool_op(EnvI& env, Ctx& ctx, const Ctx& ctx0, const Ctx& ctx1, Expres
   Expression* le0 = nullptr;
   Expression* le1 = nullptr;
 
-  if (e0.r()->type().isint() && !e0.r()->type().isOpt() && bot != BOT_IN) {
+  if (bot == BOT_IN) {
+    if (e1.r()->type().isPar() && e0.r()->type().isint() && !e0.r()->type().isOpt()) {
+      bool has_infinity = false;
+      {
+        GCLock lock;
+        IntSetVal* dom = eval_intset(env, e1.r());
+        has_infinity = dom->min(0).isMinusInfinity() || dom->max(dom->size() - 1).isPlusInfinity();
+      }
+      if (has_infinity) {
+        KeepAlive ka;
+        {
+          GCLock lock;
+          Call* call = Call::a(bo->loc(), env.constants.ids.mzn_set_in_internal, {e0.r(), e1.r()});
+          call->type(Type::varbool());
+          call->decl(env.model->matchFn(env, call, false));
+          ka = call;
+        }
+        return flat_exp(env, ctx, ka(), r, b);
+      }
+    }
+    // Otherwise translate to set_in as normal
+  } else if (e0.r()->type().isint() && !e0.r()->type().isOpt()) {
     le0 = get_linexp<IntLit>(env, e0.r());
-  } else if (e0.r()->type().isfloat() && !e0.r()->type().isOpt() && bot != BOT_IN) {
+  } else if (e0.r()->type().isfloat() && !e0.r()->type().isOpt()) {
     le0 = get_linexp<FloatLit>(env, e0.r());
   }
   if (le0 != nullptr) {
@@ -1291,20 +1316,41 @@ EE flatten_bool_op(EnvI& env, Ctx& ctx, const Ctx& ctx0, const Ctx& ctx1, Expres
       }
     }
 
+    bool needsFlatten = true;
     auto cit = env.cseMapFind(cc);
     if (cit != env.cseMapEnd()) {
-      ees[2].b = cit->second.r;
-      if (doubleNeg) {
-        Type t = ees[2].b()->type();
-        ees[2].b = new UnOp(Location().introduce(), UOT_NOT, ees[2].b());
-        ees[2].b()->type(t);
+      needsFlatten = false;
+      auto* ident = cit->second.r->dynamicCast<Id>();
+      if (ident != nullptr && ident->decl() != nullptr && ident->decl()->e() != nullptr) {
+        BCtx cseCtx;
+        bool annotated;
+        std::tie(cseCtx, annotated) = env.annToCtx(ident->decl());
+        if (cseCtx != ctx.b && cseCtx != C_ROOT && cseCtx != C_MIX) {
+          // Can't use CSE value because context doesn't match
+          needsFlatten = true;
+          if (ctx.b != C_ROOT) {
+            ctx.b = C_MIX;
+          }
+          env.addCtxAnn(ident->decl(), C_MIX);
+          env.cseMapRemove(e);
+          env.cseMapRemove(cc);
+        }
       }
-      if (Id* id = ees[2].b()->dynamicCast<Id>()) {
-        add_ctx_ann(env, id->decl(), ctx.b);
+      if (!needsFlatten) {
+        ees[2].b = cit->second.r;
+        if (doubleNeg) {
+          Type t = ees[2].b()->type();
+          ees[2].b = new UnOp(Location().introduce(), UOT_NOT, ees[2].b());
+          ees[2].b()->type(t);
+        }
+        if (Id* id = ees[2].b()->dynamicCast<Id>()) {
+          env.addCtxAnn(id->decl(), ctx.b);
+        }
+        ret.r = conj(env, r, ctx, ees);
+        GC::unlock();
       }
-      ret.r = conj(env, r, ctx, ees);
-      GC::unlock();
-    } else {
+    }
+    if (needsFlatten) {
       bool singleExp = true;
       for (auto& ee : ees) {
         if (!istrue(env, ee.b())) {
@@ -1329,7 +1375,7 @@ EE flatten_bool_op(EnvI& env, Ctx& ctx, const Ctx& ctx0, const Ctx& ctx1, Expres
           ees[2].b()->type(t);
         }
         if (Id* id = ees[2].b()->dynamicCast<Id>()) {
-          add_ctx_ann(env, id->decl(), ctx.b);
+          env.addCtxAnn(id->decl(), ctx.b);
         }
         ret.r = conj(env, r, ctx, ees);
       }
@@ -1462,7 +1508,8 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
             Expression* e_todo = todo.back();
             todo.pop_back();
             auto* e_bo = e_todo->dynamicCast<BinOp>();
-            if ((e_bo != nullptr) && e_bo->op() == (negArgs ? BOT_OR : BOT_AND)) {
+            if ((e_bo != nullptr) && e_bo->op() == (negArgs ? BOT_OR : BOT_AND) &&
+                (e_bo->decl() == nullptr || e_bo->decl()->e() == nullptr)) {
               todo.push_back(e_bo->rhs());
               todo.push_back(e_bo->lhs());
             } else {
@@ -1478,7 +1525,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
         GC::unlock();
         ret = flat_exp(env, ctx, c, r, b);
         if (Id* id = ret.r()->dynamicCast<Id>()) {
-          add_ctx_ann(env, id->decl(), ctx.b);
+          env.addCtxAnn(id->decl(), ctx.b);
         }
         break;
       }
@@ -1490,7 +1537,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
         GC::unlock();
         ret = flat_exp(env, ctx, c, r, b);
         if (Id* id = ret.r()->dynamicCast<Id>()) {
-          add_ctx_ann(env, id->decl(), ctx.b);
+          env.addCtxAnn(id->decl(), ctx.b);
         }
         break;
       }
@@ -1654,7 +1701,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
       GC::unlock();
       ret = flat_exp(env, ctx, c, r, b);
       if (Id* id = ret.r()->dynamicCast<Id>()) {
-        add_ctx_ann(env, id->decl(), ctx.b);
+        env.addCtxAnn(id->decl(), ctx.b);
       }
     } break;
     case BOT_EQUIV:
@@ -1697,7 +1744,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
         Id* id = e0.r()->cast<Id>();
         ctx1.b = C_MIX;
         (void)flat_exp(env, ctx1, boe1, id->decl(), env.constants.varTrue);
-        add_ctx_ann(env, id->decl(), ctx1.b);
+        env.addCtxAnn(id->decl(), ctx1.b);
         ret.b = bind(env, Ctx(), b, env.constants.literalTrue);
         ret.r = bind(env, Ctx(), r, env.constants.literalTrue);
         break;
