@@ -630,7 +630,7 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
           argctx.b = C_MIX;
         }
         auto* al = c->arg(0)->cast<ArrayLit>();
-        ArrayLit* al_new;
+        KeepAlive al_new;
         if (al->flat()) {
           al_new = al;
         } else {
@@ -650,12 +650,14 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
               flat_args.emplace_back(res.r());
             }
           }
-          GCLock lock;
-          al_new = new ArrayLit(al->loc(), to_exp_vec(flat_args));
-          al_new->type(Type::varbool(1));
-          al_new->flat(true);
+          {
+            GCLock lock;
+            al_new = new ArrayLit(al->loc(), to_exp_vec(flat_args));
+            al_new()->type(Type::varbool(1));
+            al_new()->cast<ArrayLit>()->flat(true);
+          }
         }
-        args_ee[0] = EE(al_new, env.constants.literalTrue);
+        args_ee[0] = EE(al_new(), env.constants.literalTrue);
       } else {
         BCtx transfer_ctx = c->type().bt() == Type::BT_INT ? nctx.i : nctx.b;
         for (unsigned int i = c->argCount(); (i--) != 0U;) {
@@ -985,12 +987,33 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
       cr_c->decl(decl);
       cr = cr_c;
     }
+
+    bool needsFlatten = true;
     auto cit = env.cseMapFind(cr());
     if (cit != env.cseMapEnd()) {
+      needsFlatten = false;
       if (e->type().isbool() && !e->type().isOpt()) {
-        ret.b = bind(env, Ctx(), b, env.constants.literalTrue);
-        args_ee.emplace_back(nullptr, cit->second.r);
-        ret.r = conj(env, r, ctx, args_ee);
+        auto* ident = cit->second.r->dynamicCast<Id>();
+        if (ident != nullptr && ident->decl() != nullptr && ident->decl()->e() != nullptr) {
+          BCtx cseCtx;
+          bool annotated;
+          std::tie(cseCtx, annotated) = env.annToCtx(ident->decl());
+          if (cseCtx != ctx.b && cseCtx != C_ROOT && cseCtx != C_MIX) {
+            // Can't use CSE value because context doesn't match
+            needsFlatten = true;
+            if (ctx.b != C_ROOT) {
+              ctx.b = C_MIX;
+            }
+            env.addCtxAnn(ident->decl(), C_MIX);
+            env.cseMapRemove(e);
+            env.cseMapRemove(cr());
+          }
+        }
+        if (!needsFlatten) {
+          ret.b = bind(env, Ctx(), b, env.constants.literalTrue);
+          args_ee.emplace_back(nullptr, cit->second.r);
+          ret.r = conj(env, r, ctx, args_ee);
+        }
       } else {
         if (env.ignorePartial) {
           ret.b = bind(env, Ctx(), b, env.constants.literalTrue);
@@ -1000,7 +1023,9 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
         }
         ret.r = bind(env, ctx, r, cit->second.r);
       }
-    } else {
+    }
+
+    if (needsFlatten) {
       for (unsigned int i = 0; i < decl->paramCount(); i++) {
         if (decl->param(i)->type().dim() > 0) {
           // Check array index sets
@@ -1039,8 +1064,8 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
                     needToConstrain = true;
                   } else {
                     IntBounds ib = compute_int_bounds(env, curArg());
-                    needToConstrain = !ib.valid || isv->empty() || ib.l < isv->min(0) ||
-                                      ib.u > isv->max(isv->size() - 1);
+                    needToConstrain =
+                        !ib.valid || isv->empty() || !isv->isSubset(IntSetVal::Range(ib.l, ib.u));
                   }
                 }
                 if (needToConstrain) {
@@ -1053,7 +1078,9 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
                     (void)flat_exp(env, Ctx(), domconstraint(), env.constants.varTrue,
                                    env.constants.varTrue);
                   } else {
-                    EE ee = flat_exp(env, Ctx(), domconstraint(), nullptr, env.constants.varTrue);
+                    Ctx domctx = ctx;
+                    domctx.neg = false;
+                    EE ee = flat_exp(env, domctx, domconstraint(), nullptr, env.constants.varTrue);
                     ee.b = ee.r;
                     args_ee.push_back(ee);
                   }
@@ -1067,8 +1094,8 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
                   needToConstrain = true;
                 } else {
                   FloatBounds fb = compute_float_bounds(env, curArg());
-                  needToConstrain = !fb.valid || fsv->empty() || fb.l < fsv->min(0) ||
-                                    fb.u > fsv->max(fsv->size() - 1);
+                  needToConstrain =
+                      !fb.valid || fsv->empty() || !fsv->isSubset(FloatSetVal::Range(fb.l, fb.u));
                 }
 
                 if (needToConstrain) {
@@ -1081,7 +1108,9 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
                     (void)flat_exp(env, Ctx(), domconstraint(), env.constants.varTrue,
                                    env.constants.varTrue);
                   } else {
-                    EE ee = flat_exp(env, Ctx(), domconstraint(), nullptr, env.constants.varTrue);
+                    Ctx domctx = ctx;
+                    domctx.neg = false;
+                    EE ee = flat_exp(env, domctx, domconstraint(), nullptr, env.constants.varTrue);
                     ee.b = ee.r;
                     args_ee.push_back(ee);
                   }
@@ -1129,7 +1158,7 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
           if (r == nullptr || (r != nullptr && r->e() != nullptr)) {
             reif_b = new_vardecl(env, Ctx(), new TypeInst(Location().introduce(), Type::varbool()),
                                  nullptr, nullptr, nullptr);
-            add_ctx_ann(env, reif_b, ctx.b);
+            env.addCtxAnn(reif_b, ctx.b);
             if (reif_b->ti()->domain() != nullptr) {
               if (reif_b->ti()->domain() == env.constants.literalTrue) {
                 bind(env, ctx, r, env.constants.literalTrue);
@@ -1214,7 +1243,7 @@ EE flatten_call(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, VarD
           args_ee.push_back(res);
           ret.b = conj(env, b, Ctx(), args_ee);
           add_path_annotation(env, res.r());
-          ret.r = bind(env, ctx, r, res.r());
+          ret.r = bind(env, Ctx(), r, res.r());
           if (!ctx.neg && !cr_c->type().isAnn()) {
             env.cseMapInsert(cr_c, ret);
           }
