@@ -783,17 +783,9 @@ void optimize(Env& env, bool chain_compression) {
               }
               if (auto* vdi = item->dynamicCast<VarDeclI>()) {
                 // The variable occurs in the RHS of another variable, so
-                // if that is an array variable, simplify all constraints that
-                // mention the array variable
-                if ((vdi->e()->e() != nullptr) && vdi->e()->e()->isa<ArrayLit>()) {
-                  auto ait = envi.varOccurrences.itemMap.find(vdi->e()->id()->decl()->id());
-                  if (ait.first) {
-                    for (auto* aitem : *ait.second) {
-                      simplify_bool_constraint(envi, aitem, vd, remove, vardeclQueue,
-                                               constraintQueue, toRemove, deletedVarDecls,
-                                               nonFixedLiteralCount);
-                    }
-                  }
+                // if that is an array variable, push it onto the stack for processing
+                if (vdi->e()->e() != nullptr && vdi->e()->e()->isa<ArrayLit>()) {
+                  push_vardecl(envi, envi.varOccurrences.idx.get(vdi->e()->id()), vardeclQueue);
                   continue;
                 }
               }
@@ -862,9 +854,7 @@ void optimize(Env& env, bool chain_compression) {
             // push all constraints that depend on the array
             substitute_fixed_vars(envi, item, deletedVarDecls);
             push_dependent_constraints(envi, item->cast<VarDeclI>()->e()->id(), constraintQueue);
-          } else if ((c == nullptr) || !(c->id() == envi.constants.ids.forall ||
-                                         c->id() == envi.constants.ids.exists ||
-                                         c->id() == envi.constants.ids.clause)) {
+          } else {
             // For any constraint that is not forall, exists or clause,
             // substitute fixed arguments, then simplify it
             substitute_fixed_vars(envi, item, deletedVarDecls);
@@ -1331,7 +1321,7 @@ bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarD
         BinOpType bot =
             c->arg(0)->isa<Id>() ? (is_true ? BOT_LQ : BOT_GR) : (is_true ? BOT_GQ : BOT_LE);
         IntSetVal* newDomain = LinearTraits<IntLit>::limitDomain(bot, domain, eval_int(env, arg));
-        if (newDomain->card() == 0) {
+        if (newDomain->empty()) {
           env.fail();
         } else {
           ident->decl()->ti()->domain(new SetLit(Location().introduce(), newDomain));
@@ -1376,6 +1366,7 @@ bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarD
       if (vd != nullptr) {
         if (vd->ti()->domain() != nullptr) {
           vd_dom = eval_intset(env, vd->ti()->domain());
+          assert(!vd_dom->empty());
           if (vd_dom->max() < 0 || vd_dom->min() > 1) {
             env.fail();
             return true;
@@ -1599,7 +1590,8 @@ int decrement_non_fixed_vars(std::unordered_map<Expression*, int>& nonFixedLiter
       auto* al = follow_id(c->arg(i))->cast<ArrayLit>();
       nonFixedVars += static_cast<int>(al->size());
       for (unsigned int j = al->size(); (j--) != 0U;) {
-        if ((*al)[j]->type().isPar()) {
+        if ((*al)[j]->type().isPar() ||
+            ((*al)[j]->isa<Id>() && (*al)[j]->cast<Id>()->decl()->ti()->domain() != nullptr)) {
           nonFixedVars--;
         }
       }
